import cgi
import codecs
import typing
from dataclasses import dataclass, field

from parsel import Selector, SelectorList
from furl import furl


def is_known_encoding(encoding: str) -> bool:
    """
    Return `True` if `encoding` is a known codec.
    """
    try:
        codecs.lookup(encoding)
    except LookupError:
        return False
    return True


class TextDecoder:
    """
    Handles incrementally decoding bytes into text
    """

    def __init__(self, encoding: typing.Optional[str] = None):
        self.decoder: typing.Optional[codecs.IncrementalDecoder] = None
        if encoding is not None:
            self.decoder = codecs.getincrementaldecoder(encoding)(errors="replace")

    def decode(self, data: bytes) -> str:
        if self.decoder is None:
            # If this is the first decode pass then we need to determine which
            # encoding to use by attempting UTF-8 and raising any decode errors.
            attempt_utf_8 = codecs.getincrementaldecoder("utf-8")(errors="strict")
            try:
                attempt_utf_8.decode(data)
            except UnicodeDecodeError:
                # Could not decode as UTF-8. Use Windows 1252.
                self.decoder = codecs.getincrementaldecoder("cp1252")(errors="replace")
            else:
                # Can decode as UTF-8. Use UTF-8 with lenient error settings.
                self.decoder = codecs.getincrementaldecoder("utf-8")(errors="replace")

        return self.decoder.decode(data)

    def flush(self) -> str:
        if self.decoder is None:
            return ""
        return self.decoder.decode(b"", True)


class MySelector(Selector):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.selectorlist_cls = MySelectorList

    def css(self, query, domain=None):
        result = super(MySelector, self).css(query)
        result.domain = domain
        return result

    def xpath(self, query, namespaces=None, domain=None, **kwargs):
        result = super(MySelector, self).xpath(query, namespaces=namespaces, **kwargs)
        result.domain = domain
        return result


class MySelectorList(SelectorList):
    domain: str = ''

    def proc_result(self, value, domain, is_url):
        if is_url and domain and value.startswith('/'):
            value = domain + value
        if type(value) == str:
            value = value.strip()
        return value

    def getall(self, is_url=False):
        result = super(MySelectorList, self).getall()
        return [self.proc_result(x, domain=self.domain, is_url=is_url) for x in result]

    def get(self, default=None, is_url=False):
        result = super(MySelectorList, self).get(default=default)
        result = self.proc_result(result, self.domain, is_url)
        return result


@dataclass
class Response(object):
    ok: bool = field(default=False)

    url: str = field(default_factory=str)
    metadata: dict = field(default_factory=dict)
    content: bytes = field(default_factory=bytes)
    status_code: int = field(default_factory=int)
    history: list = field(default_factory=list)
    headers: dict = field(default_factory=dict)
    cookies: dict = field(default_factory=dict)

    selector: MySelector = field(default=None, init=False)
    domain: str = field(default_factory=str, init=False)
    host: str = field(default_factory=str, init=False)
    _text: str = field(default_factory=str, init=False)
    _encoding: str = field(default=None, init=False)

    def __post_init__(self):
        self.selector = MySelector(self.text)
        self.domain = furl(self.url).origin
        self.host = furl(self.url).host

    @property
    def charset_encoding(self) -> typing.Optional[str]:
        """
        Return the encoding, as specified by the Content-Type header.
        """
        content_type = self.headers.get("Content-Type")
        if content_type is None:
            return None

        _, params = cgi.parse_header(content_type)
        if "charset" not in params:
            return None

        return params["charset"].strip("'\"")

    @property
    def encoding(self) -> typing.Optional[str]:
        """
        Return the encoding, which may have been set explicitly, or may have
        been specified by the Content-Type header.
        """
        if not self._encoding:
            encoding = self.charset_encoding
            if encoding is None or not is_known_encoding(encoding):
                self._encoding = "utf-8"
            else:
                self._encoding = encoding
        return self._encoding

    @encoding.setter
    def encoding(self, value: str) -> None:
        self._encoding = value

    @property
    def text(self) -> str:
        if not self._text:
            content = self.content
            if not content:
                self._text = ""
            else:
                decoder = TextDecoder(encoding=self.encoding)
                self._text = "".join([decoder.decode(self.content), decoder.flush()])
        return self._text

    def json(self, **kwargs):
        pass

    def xpath(self, query):
        return self.selector.xpath(query, domain=self.domain)

    def css(self, query):
        return self.selector.css(query, domain=self.domain)

    def to_url(self, urls: typing.Union[list, str]):
        urls = urls if type(urls) == list else [urls]
        urls = [furl(self.url).join(url).url for url in urls]
        return urls

    def re(self, regex, replace_entities=True):
        return self.selector.re(regex, replace_entities=replace_entities)

    def re_first(self, regex, default=None, replace_entities=True):
        return self.selector.re_first(regex, default=default, replace_entities=replace_entities)



import asyncio
import os
import sys
import weakref
from dataclasses import field
from datetime import datetime
from functools import reduce
from inspect import isawaitable
from pathlib import Path
from types import AsyncGeneratorType
from typing import Optional, Union, Iterable, Callable, Tuple, Coroutine

from fake_useragent import UserAgent
from httpx import AsyncClient
from loguru import logger
from pyppeteer import launch
from pyppeteer.browser import Browser

from hssp.exceptions import NotImplementedParseError, SpiderHookError, InvalidCallbackResult
from hssp.item import Item, ItemMgr
from hssp.middleware import Middleware
from hssp.request import Request
from hssp.response import Response
from hssp.settings import Setting, ReqSetting

if sys.version_info >= (3, 8) and sys.platform.startswith("win"):
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

if sys.version_info >= (3, 9):
    async_all_tasks = asyncio.all_tasks
    async_current_task = asyncio.current_task
else:
    async_all_tasks = asyncio.Task.all_tasks
    async_current_task = asyncio.tasks.Task.current_task
try:
    import uvloop

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    uvloop = None


class Spider:
    name: str = 'hssp'

    # 开始的url列表
    start_urls: list = None

    # 向第一个回调函数中传递的数据
    metadata: dict = field(default={})

    # 统计成功和失败的个数
    _success_counts: int = 0
    _failed_counts: int = 0

    # 任务队列
    _worker_tasks: list = []

    _fake_useragent: UserAgent = None

    _browser: Browser = None

    def __init__(
            self,
            setting: Setting = None,
            loop: Optional[asyncio.AbstractEventLoop] = None,
            middleware: Union[Iterable, Middleware] = None,
            **kwargs
    ):
        # 爬虫的设置类
        self._setting = setting or Setting()

        # 事件循环
        self._loop = loop

        # 中间件
        if isinstance(middleware, list):
            self._middleware = reduce(lambda x, y: x + y, middleware)
        else:
            self._middleware = middleware or Middleware()

        # 设置事件循环
        asyncio.set_event_loop(self._loop)

        # 异步队列
        self.request_queue = asyncio.Queue()

        # 信号量，用于控制并发
        self._sem = asyncio.Semaphore(self._setting.concurrency)

        # 日志
        logger.remove()
        logger.add(sys.stderr, level=self._setting.log_Level)
        self.logger = logger

        # 请求会话
        self.session = AsyncClient(proxies=self._setting.proxies, verify=False)

        # 请求的全局设置
        self.req_setting = self._setting
        self.req_setting.__class__ = ReqSetting

        # 设置假ua
        if self._setting.random_ua:
            fake_useragent_path = Path(os.path.abspath(__file__)).parent / 'res/fake_useragent_0.1.11.json'
            self._fake_useragent = UserAgent(path=str(fake_useragent_path), verify_ssl=False)

        if isinstance(self._setting.out_path, list):
            outpus = self._setting.out_path
        else:
            outpus = [self._setting.out_path]
        self.item_mgr = ItemMgr(outputs=outpus)

    async def __async_init__(self):
        """
        异步初始化方法
        :return:
        """
        self._browser = await launch(
            autoClose=False,
            handleSIGINT=False,
            handleSIGTERM=False,
            headless=False,
            ignoreHTTPSErrors=True,
            ignoreDefaultArgs=['--enable-automation'],
            userDataDir=str(Path(os.environ.get('APPDATA')) / "hssp_user_data"),
            args=[
                '--disable-infobars',
                '--no-sandbox'
            ]
        )

    async def process_failed_response(self, request, response):
        """
        响应失败的处理
        :param request: 请求
        :param response: 响应
        :return:
        """
        pass

    async def process_succeed_response(self, request, response):
        """
        响应成功的处理
        :param request: 请求
        :param response: 响应
        :return:
        """
        pass

    async def process_item(self, item: Item):
        """
        处理item
        :param item: Item
        :return:
        """
        self.item_mgr.add_item(item)
        if not self._setting.out_path:
            return
        await self.item_mgr.export()

    async def parse(self, response):
        """
        字类必须实现。用于处理start_urls的回调
        :param response: Response
        :return:
        """
        raise NotImplementedParseError("<!!! parse function is expected !!!>")

    async def _run_spider_hook(self, hook_func):
        """
        运行爬虫开始之后、结束之前的函数
        :param hook_func: aws function
        :return:
        """
        if not callable(hook_func):
            return
        try:
            aws_hook_func = hook_func(weakref.proxy(self))
            if isawaitable(aws_hook_func):
                await aws_hook_func
        except Exception as e:
            raise SpiderHookError(f"<Hook {hook_func.__name__}: {e}")

    async def process_start_urls(self):
        """
        处理开始的地址
        :return: 异步请求迭代器
        """
        for url in self.start_urls:
            yield await self.get(url=url, callback=self.parse, metadata=self.metadata)

    async def _process_response(self, request: Request, response: Response):
        """
        响应成功或失败的处理
        :param request: 请求
        :param response: 响应
        :return:
        """
        if response.ok:
            # 处理成功的响应
            self._success_counts += 1
            await self.process_succeed_response(request, response)
        else:
            # 处理失败的响应
            self._failed_counts += 1
            await self.process_failed_response(request, response)

    async def _run_request_middleware(self, request: Request):
        """
        运行请求中间件
        :param request: 请求
        :return:
        """
        if not self._middleware.request_middleware:
            return
        for middleware in self._middleware.request_middleware:
            if not callable(middleware):
                continue
            try:
                aws_middleware_func = middleware(self, request)
                if isawaitable(aws_middleware_func):
                    await aws_middleware_func
                else:
                    self.logger.error(f"<Middleware {middleware.__name__}: 必须是一个协程函数")
            except Exception as e:
                self.logger.error(f"<Middleware {middleware.__name__}: {e}")

    async def _run_response_middleware(self, request: Request, response: Response):
        """
        运行响应中间件
        :param request: 请求
        :param response: 响应
        :return:
        """
        if not self._middleware.response_middleware:
            return
        for middleware in self._middleware.response_middleware:
            if not callable(middleware):
                continue
            try:
                aws_middleware_func = middleware(self, request, response)
                if isawaitable(aws_middleware_func):
                    await aws_middleware_func
                else:
                    self.logger.error(f"<Middleware {middleware.__name__}: 必须是一个协程函数")
            except Exception as e:
                self.logger.error(f"<Middleware {middleware.__name__}: {e}")

    async def get(
            self,
            url: str,
            *,
            params: dict = None,
            headers: dict = None,
            cookies: dict = None,
            callback: Callable = None,
            metadata: dict = None,
            is_browser: bool = False,
            **kwargs
    ) -> Request:
        """
        发生一个get请求
        :param url: 请求地址
        :param params: 地址中带的参数
        :param headers: 请求头
        :param cookies: 传递的cookies
        :param callback: 请求结束之后的回调函数
        :param metadata: 向函数中传递的数据
        :param is_browser: 是否使用浏览器获取，速度会慢，但会执行网页中的js脚本
        :param kwargs: httpx 请求的其他参数
        :return:
        """
        ua = self._fake_useragent.random if self._setting.random_ua else None
        return Request(
            "GET",
            url,
            self.req_setting,
            params=params,
            user_agent=ua,
            headers=headers,
            cookies=cookies,
            callback=callback,
            metadata=metadata,
            browser=self._browser,
            is_browser=is_browser,
            **kwargs
        )

    async def post(
            self,
            url: str,
            *,
            data: dict = None,
            json: dict = None,
            params: dict = None,
            headers: dict = None,
            cookies: dict = None,
            callback: Callable = None,
            metadata: dict = None,
            **kwargs
    ) -> Request:
        """
        发送一个post请求
        :param url: 请求地址
        :param data: 传递的数据
        :param json: 传递的json
        :param params: 地址中带的参数
        :param headers: 请求头
        :param cookies: 传递的cookies
        :param callback: 请求结束之后的回调函数
        :param metadata: 向函数中传递的数据
        :param kwargs: httpx 请求的其他参数
        :return:
        """
        ua = self._fake_useragent.random if self._setting.random_ua else None
        return Request(
            "POST",
            url,
            self.req_setting,
            data=data,
            json=json,
            params=params,
            user_agent=ua,
            headers=headers,
            cookies=cookies,
            callback=callback,
            metadata=metadata,
            **kwargs
        )

    async def handle_request(self, request: Request) -> Tuple[AsyncGeneratorType, Response]:
        """
        处理请求
        :param request:
        :return:
        """
        callback_result, response = None, None
        try:
            await self._run_request_middleware(request)
            callback_result, response = await request.request(self._sem)
            await self._run_response_middleware(request, response)
            await self._process_response(request=request, response=response)
        except NotImplementedParseError as e:
            self.logger.error(e)
            await self.stop()
        except Exception as e:
            self.logger.error(f"<Callback[{request.callback.__name__}]: [{e.__class__.__name__}]{e}")
        finally:
            await request.close()

        return callback_result, response

    @classmethod
    def start(
            cls,
            setting: Setting = None,
            middleware: Union[Iterable, Middleware] = None,
            loop: Optional[asyncio.AbstractEventLoop] = None,
            after_start=None,
            before_stop=None,
            close_event_loop=True,
            **kwargs
    ):
        """
        类方法。 开始爬虫
        :param setting: 设置
        :param after_start: 处理开始之后
        :param before_stop: 处理结束之前
        :param middleware: 自定义中间件，可以的列表或单独中间件
        :param loop: 事件循环
        :param close_event_loop: 结束后是否关闭事件循环
        :return: An instance of :cls:`Spider`
        """
        loop = loop or asyncio.new_event_loop()
        spider_ins = cls(setting=setting, middleware=middleware, loop=loop, **kwargs)

        # Actually start crawling
        spider_ins._loop.run_until_complete(
            spider_ins._start(after_start=after_start, before_stop=before_stop)
        )
        spider_ins._loop.run_until_complete(spider_ins._loop.shutdown_asyncgens())
        if close_event_loop:
            spider_ins._loop.close()

        return spider_ins

    async def _start(self, after_start=None, before_stop=None):
        """
        开始爬虫的内部函数
        :param after_start: 处理开始之后
        :param before_stop: 处理结束之前
        :return:
        """
        self.logger.info(f"{self.name} Spider started!")
        # await self.__async_init__()

        start_time = datetime.now()

        # 运行爬虫开始之前的hook函数
        await self._run_spider_hook(after_start)

        # Actually run crawling
        try:
            await self.start_master()
        finally:
            # 运行爬虫结束后的hook函数
            await self._run_spider_hook(before_stop)

            await self.stop()
            # Display logs about this crawl task
            end_time = datetime.now()
            self.logger.success(f"总请求数量: {self._failed_counts + self._success_counts}")

            if self._failed_counts:
                self.logger.success(f"失败的请求: {self._failed_counts}")
            self.logger.success(f"总用时: {end_time - start_time}")
            self.logger.success("爬取完成")

    async def start_master(self):
        """
        开始爬取
        """
        async for request_ins in self.process_start_urls():
            self.request_queue.put_nowait(self.handle_request(request_ins))
        workers = [asyncio.ensure_future(self.start_worker()) for _ in range(self._setting.concurrency)]
        for worker in workers:
            self.logger.info(f"Worker started: {id(worker)}")
        await self.request_queue.join()

    async def start_worker(self):
        """
        开始爬取工作
        :return:
        """
        while True:
            request_item = await self.request_queue.get()
            self._worker_tasks.append(request_item)
            if self.request_queue.empty():
                results = await asyncio.gather(*self._worker_tasks, return_exceptions=True)
                for task_result in results:
                    if not isinstance(task_result, RuntimeError) and task_result:
                        callback_results, response = task_result
                        if isinstance(callback_results, AsyncGeneratorType):
                            await self._process_async_callback(callback_results)
                self._worker_tasks = []
            self.request_queue.task_done()

    async def _process_async_callback(self, callback_results: AsyncGeneratorType):
        try:
            async for callback_result in callback_results:
                # 处理异步
                if isinstance(callback_result, Coroutine):
                    callback_result = await callback_result

                # 处理异步生成器类型：再次调用
                if isinstance(callback_result, AsyncGeneratorType):
                    await self._process_async_callback(callback_result)

                # 处理下一个请求
                elif isinstance(callback_result, Request):
                    self.request_queue.put_nowait(self.handle_request(request=callback_result))

                # 处理item
                elif isinstance(callback_result, Item):
                    # Process target item
                    await self.process_item(callback_result)

                # 其他的就报错
                else:
                    callback_result_name = type(callback_result).__name__
                    raise InvalidCallbackResult(f"<Parse invalid callback result type: {callback_result_name}>")
        except Exception as e:
            self.logger.exception(e)

    async def stop(self):
        """
        完成正在运行的任务，取消剩余的任务
        :return:
        """
        self.logger.success(f"Stopping spider: {self.name}")
        await self.session.aclose()
        await self.cancel_all_tasks()

    @staticmethod
    async def cancel_all_tasks():
        """
        取消所有的任务
        :return:
        """
        tasks = []
        for task in async_all_tasks():
            if task is not async_current_task():
                tasks.append(task)
                task.cancel()
        await asyncio.gather(*tasks, return_exceptions=True)



class IgnoreThisItem(Exception):
    """
    忽略这个Item
    """
    pass


class InvalidCallbackResult(Exception):
    """
    回调结果无效
    """
    pass


class NotImplementedParseError(Exception):
    """
    一个解析函数都没有
    """
    pass


class SpiderHookError(Exception):
    """
    在爬虫执行Hook的时候发生异常
    """
    pass


class BrowerError(Exception):
    """
    使用浏览器请求时发生异常
    """



from dataclasses import dataclass, field
# from enum import Enum, auto
from typing import Union, Dict, Iterable

from httpx import AsyncClient


@dataclass
class BaseSetting:
    pass


@dataclass
class ReqSetting(BaseSetting):
    user_agent: str = field(default="")
    encoding: str = field(default="")
    headers: dict = field(default_factory=dict)
    cookies: dict = field(default_factory=dict)
    proxies: Union[Dict, str] = field(default=None)
    timeout: float = field(default=10)
    delay: int = field(default=0)
    retrys_count: int = field(default=10)
    retries_delay: int = field(default=0)
    allow_codes: list = field(default=None)
    session: AsyncClient = field(default=None)
    random_ua: bool = field(default=False)


@dataclass
class SpSetting(BaseSetting):
    concurrency: int = field(default=64)
    log_Level: str = field(default="DEBUG")
    out_path: Union[Iterable, str] = field(default_factory=str)


@dataclass
class Setting(SpSetting, ReqSetting):
    pass



import asyncio
import functools
import json
import traceback
from dataclasses import dataclass, field
from inspect import iscoroutinefunction
from types import AsyncGeneratorType
from typing import Callable, Tuple

from furl import furl
from hssp.exceptions import BrowerError
from hssp.response import Response
from hssp.settings import ReqSetting
from httpx import AsyncClient, Timeout
from loguru import logger
from pyppeteer.browser import Browser
from pyppeteer.network_manager import Request as PYPPRequest
from pyppeteer.page import Page
from tenacity import (
    AsyncRetrying, Future, RetryCallState,
    stop_after_attempt, wait_fixed, wait_random
)

try:
    import uvloop

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    uvloop = None


@dataclass
class Request(object):
    method: str
    url: str
    _setting: ReqSetting
    callback: Callable = field(default=None)
    metadata: dict = field(default=None)
    data: dict = field(default=None)
    json: dict = field(default=None)
    params: dict = field(default=None)
    user_agent: str = field(default=None)
    headers: dict = field(default=None)
    cookies: dict = field(default=None)
    timeout: float = field(default=None)
    session: AsyncClient = field(default=None)
    browser: Browser = field(default=None)
    is_browser: bool = field(default=False)
    _page: Page = field(default=None, init=False)
    kwargs: dict = field(default_factory=dict)

    def __post_init__(self):
        # 设置user_agent
        self.user_agent = self.user_agent or self._setting.user_agent

        # 设置httpx会话
        if not self.session:
            self.session = AsyncClient(proxies=self._setting.proxies, verify=False)

        # 设置session的headers、cookies、timeout
        self.session.headers = self.headers or self._setting.headers
        self.session.cookies = self.cookies or self._setting.cookies
        self.session.timeout = Timeout(self.timeout or self._setting.timeout)

        # user_agent设置到headers中
        if self.user_agent:
            self.session.headers.update({"user-agent": self.user_agent})

        self.headers = dict(self.session.headers)

        # 设置允许的状态码 默认是200-300之间
        allow_codes = list(range(200, 300))
        if not self._setting.allow_codes:
            self._setting.allow_codes = allow_codes
        else:
            self._setting.allow_codes.extend(allow_codes)

    async def close(self):
        """
        关闭session
        :return:
        """
        await self.session.aclose()

    def retry_handler(self, retry_state: RetryCallState):
        """
        处理重试之后和重试失败
        :param retry_state:
        :return:
        """
        outcome: Future = retry_state.outcome
        attempt_number = retry_state.attempt_number
        exception_type = type(outcome.exception()).__name__
        exception_msg = list(outcome.exception().args)
        stack_msg = traceback.format_stack()[2].strip()
        log_msg = f"[{self.method}] {self.url} 发生异常: [{exception_type}]: {exception_msg}"

        if 'retry_error_callback' in stack_msg:
            logger.error(f"{log_msg} {attempt_number}次重试全部失败")
            return None, Response(url=self.url, metadata=self.metadata)
        else:
            logger.error(f"{log_msg} 第{attempt_number}次重试")

        if self._page:
            asyncio.ensure_future(self._page.close())

    async def request(self, sem: asyncio.Semaphore) -> Tuple[AsyncGeneratorType, Response]:
        """
        使用httpx发起请求 带重试机制
        :param sem: 信号量 用于控制并发
        :return: 返回异步回调函数的执行结果和响应
        """
        if self._setting.retrys_count < 1:
            return await self._request(sem)

        # 设置重试的等候时间
        if self._setting.retries_delay:
            wait = wait_fixed(self._setting.retries_delay) + wait_random(0, 2)
        else:
            wait = wait_fixed(0)

        # 异步重试
        r = AsyncRetrying(
            stop=stop_after_attempt(self._setting.retrys_count),
            after=self.retry_handler,
            retry_error_callback=self.retry_handler,
            wait=wait
        )
        resp = await r.wraps(self._request)(sem)
        return resp

    async def _request_httpx(self, sem: asyncio.Semaphore) -> Response:
        """
        使用httpx发起请求
        :param sem: 信号量 用于控制并发
        :return:
        """
        async with sem:
            resp = await self.session.request(
                self.method,
                self.url,
                data=self.data,
                json=self.json,
                params=self.params,
                **self.kwargs
            )
        response = Response(
            ok=resp.status_code in self._setting.allow_codes,
            url=str(resp.url),
            metadata=self.metadata,
            content=resp.content,
            status_code=resp.status_code,
            history=resp.history,
            cookies=dict(resp.cookies),
            headers=resp.headers
        )
        response.encoding = self._setting.encoding or resp.encoding
        response.json = resp.json
        return response

    async def _request_brower(self, sem: asyncio.Semaphore) -> Response:
        """
        使用pyppeteer的方式打开请求数据
        :param sem: 信号量 用于控制并发
        :return:
        """

        async def set_brower_request(pypp_req: PYPPRequest):
            if pypp_req.resourceType in ['image', 'media', 'eventsource', 'websocket']:
                await pypp_req.abort()
            else:
                await pypp_req.continue_({
                    "url": url,
                    "method": self.method,
                    "headers": headers
                })

        url = furl(self.url).add(self.params or {}).url
        headers = dict(self.session.headers)
        cookies = dict(self.session.cookies)

        async with sem:
            # 打开新页面
            self._page = await self.browser.newPage()
            # 启用拦截器
            await self._page.setRequestInterception(True)
            # 更改请求
            self._page.on("request", lambda pypp_req: asyncio.ensure_future(set_brower_request(pypp_req)))

            for k, v in cookies.items():
                await self._page.setCookie({
                    "name": k,
                    "value": v,
                    "domain": furl(url).host
                })

            # 打开网页
            resp = await self._page.goto(
                url=url,
                timeout=self.session.timeout.read * 1000,
                waitUntil="networkidle2",
            )

        if not resp:
            raise BrowerError()

        content = await resp.buffer()
        content = content.encode('utf-8') if isinstance(content, str) else content
        cookies = await self._page.cookies()
        cookies = {data['name']: data['value'] for data in cookies}
        response = Response(
            ok=resp.ok,
            url=resp.url,
            metadata=self.metadata,
            content=content,
            status_code=resp.status,
            cookies=cookies,
            headers=resp.headers
        )
        await self._page.close()
        self.session.cookies = cookies
        response.json = functools.partial(json.loads, response.text)
        return response

    async def _request(self, sem: asyncio.Semaphore) -> Tuple[AsyncGeneratorType, Response]:
        """
        使用httpx发起请求
        :param sem: 信号量 用于控制并发
        :return: 返回异步回调函数的执行结果和响应
        """
        if self._setting.delay > 0:
            await asyncio.sleep(self._setting.delay)

        logger.debug(f"[{self.method}] {self.url}")

        if self.is_browser:
            response = await self._request_brower(sem)
        else:
            response = await self._request_httpx(sem)

        if self.callback and response.ok:
            if iscoroutinefunction(self.callback):
                callback_result = await self.callback(response)
            else:
                callback_result = self.callback(response)
        else:
            callback_result = None

        return callback_result, response



from pathlib import Path

import orjson
import aiofiles
from pandas import DataFrame
from pandas import concat

from hssp.item import Item


class ItemMgr(object):
    def __init__(self, outputs: list):
        self.outputs = set(outputs)
        self.df = DataFrame()
        self._data_list = []

    def add_item(self, item: Item):
        item_copy = dict()
        self._data_list.append(item.copy())
        for k, v in item.copy().items():
            if type(v) == list:
                item_copy[k] = ' |$| '.join(v)
            else:
                item_copy[k] = v
        df = DataFrame(item_copy, index=[0])
        self.df = concat([self.df, df], ignore_index=True)

    async def export(self):
        for path in self.outputs:
            suffix = Path(path).suffix
            method = getattr(self, f"to_{suffix[1:]}")
            if method:
                await method(path)

    async def to_json(self, path):
        """
        导出至json文件
        Args:
            path: 文件路径

        Returns:

        """
        # 使用orjson 方式导出 aiofiles 写入文件
        async with aiofiles.open(path, mode='wt', encoding='utf-8') as f:
            await f.write(str(orjson.dumps(self._data_list), encoding='utf-8'))

        # pandas 方式导出 暂时不用
        # return self.df.to_json(path, force_ascii=False, orient='records')

    async def to_xml(self, path):
        """
        导出至xml文件
        Args:
            path: 文件路径

        Returns:

        """
        self.df.to_xml(path, index=False)

    async def to_csv(self, path):
        """
        导出至csv文件
        Args:
            path: 文件路径

        Returns:

        """
        return self.df.to_csv(path, index=False)

    async def to_xls(self, path):
        """
        导出xls格式的excel文件
        Args:
            path: 文件路径

        Returns:

        """
        self.df.to_excel(path + 'x', index=False)

    async def to_xlsx(self, path):
        """
        导出xlsx格式的excel文件
        Args:
            path: 文件路径

        Returns:

        """
        self.df.to_excel(path, index=False)

    def to_data(self):
        """
        返回原数据
        Returns:

        """
        return self._data_list



from itertools import groupby
from operator import itemgetter
from pathlib import Path

from hssp import Spider, Response
from hssp.item import Item, Field


class WenZhenItem(Item):
    parent = Field()
    title = Field(css_select='.book-detail-title::text')
    content = Field(xpath_select='//div[@class="book-detail-content"]//text()', many=True)

    def _clean_content(self, content: list):
        return '\n'.join(content)


class WenZhen(Spider):
    start_urls = ['https://www.iwzbz.com/artical/pcbook/v2/3_1_4.html']

    async def parse(self, response: Response):
        catalog_dict = {}
        zong = response.css('#catalog a[target="_blank"]')
        zong_name = zong.css('p::text').get()
        zong_urls = zong.css('::attr(href)').get()
        catalog_dict[zong_name] = response.to_url(zong_urls)

        for child in response.css('a[href="#"][class="catalogitem"]'):
            name = child.css("::text").get()
            vid = child.css('::attr(val)').get()
            urls = response.css(f"#childitem{vid} a::attr(href)").getall()
            urls = response.to_url(urls)
            catalog_dict[name] = urls

        for name, urls in catalog_dict.items():
            for url in urls:
                yield self.get(url, callback=self.parse_content, metadata={"parent": name})

    async def parse_content(self, response: Response):
        parent = response.metadata['parent']
        item = await WenZhenItem.extract(response=response)
        item.parent = parent
        yield item


def save_md(content: str, name: str):
    content = content.replace('\n\n', '\n')
    path = Path('D:\\Users\\Documents\\穷通宝鉴') / f"{name}.md"
    path.parent.mkdir(exist_ok=True)
    with path.open('wt', encoding='utf-8') as f:
        f.write(content)


def main():
    result = WenZhen.start().item_mgr.to_data()

    for k, v in groupby(result, itemgetter('parent')):  # type: str, list
        content = f"# {k}\n"
        content += '\n'.join([f"## {data['title']}\n{data['content']}" for data in v])
        save_md(content, k)


if __name__ == '__main__':
    main()



from abc import ABCMeta
from inspect import isawaitable
from typing import Dict, Any
from copy import deepcopy
from pprint import pformat
from dataclasses import dataclass, field
from collections.abc import MutableMapping

from parsel import Selector

from hssp import Response


@dataclass()
class Field(object):
    css_select: str = field(default_factory=str)
    xpath_select: str = field(default_factory=str)
    re_select: str = field(default_factory=str)
    default: Any = field(default=None)
    is_url: bool = field(default=False)
    many: bool = field(default=False)


class ItemMeta(ABCMeta):

    def __new__(mcs, class_name, bases, attrs):
        classcell = attrs.pop('__classcell__', None)
        new_bases = tuple(getattr(base, '_class') for base in bases if hasattr(base, '_class'))
        _class = super().__new__(mcs, 'x_' + class_name, new_bases, attrs)

        fields = getattr(_class, 'fields', {})
        new_attrs = {}
        for n in dir(_class):
            v = getattr(_class, n)
            if isinstance(v, Field):
                fields[n] = v
            elif n in attrs:
                new_attrs[n] = attrs[n]

        new_attrs['fields'] = fields
        new_attrs['_class'] = _class
        if classcell is not None:
            new_attrs['__classcell__'] = classcell
        return super().__new__(mcs, class_name, bases, new_attrs)


class DictItem(MutableMapping):
    fields: Dict[str, Field] = {}

    def __init__(self, *args, **kwargs):
        self._values = {}
        if args or kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v

    def __getitem__(self, key):
        return self._values.get(key)

    def __setitem__(self, key, value):
        if key in self.fields:
            self._values[key] = value
        else:
            raise KeyError(f"{self.__class__.__name__} does not support field: {key}")

    def __delitem__(self, key):
        del self._values[key]

    def __getattr__(self, name):
        if name in self.fields:
            return self.__getitem__(name)
        if name.endswith('__field'):
            name = 'url'.split('__field')[0]
            return self.fields.get(name)
        return super().__getattribute__(name)

    def __setattr__(self, name, value):
        if name in self.fields:
            self.__setitem__(name, value)
        else:
            super().__setattr__(name, value)

    def __len__(self):
        return len(self._values)

    def __iter__(self):
        return iter(self._values)

    def keys(self):
        return self._values.keys()

    def __repr__(self):
        return pformat(self.to_dict())

    def to_dict(self):
        return dict(self)

    def copy(self):
        """Return a :func:`~copy.deepcopy` of this item.
        """
        return deepcopy(self.to_dict())


class Item(DictItem, metaclass=ItemMeta):
    target_item = Field()

    @classmethod
    async def extracts(
            cls,
            *,
            response: Response = None,
            selector: Selector = None,
            domain: str = None
    ):
        if not response and not selector:
            return
        selector = selector or response.selector
        domain = domain or response.domain or ''

        items_field = cls.fields.get('target_item', None)
        if not items_field:
            raise ValueError("<Item: target_item 不存在, 请使用 extract>")

        items_field.many = True
        items_selector = await cls.extract(response=response, selector=selector, domain=domain, target=True)
        if items_selector:
            for each_selector in items_selector:
                _item = await cls.extract(response=response, selector=each_selector, domain=domain)
                yield _item
        else:
            value_error_info = "<Item: 无法获取target_item的值>"
            raise ValueError(value_error_info)

    @classmethod
    async def extract(
            cls,
            *,
            response: Response = None,
            selector: Selector = None,
            domain: str = None,
            target=False
    ):
        if not response and not selector:
            return
        selector = selector or response.selector
        domain = domain or response.domain or ''

        _item = cls()
        for key, f in _item.fields.items():
            # target_item 判断
            if target and key == "target_item":
                return _item._extract_field(f, selector, is_source=True)
            if (target and key != "target_item") or (not target and key == "target_item"):
                continue

            # 提取字段值
            value = _item._extract_field(f, selector)
            if not value:
                continue
            # 字段后处理
            clean_method = getattr(_item, f"_clean_{key}", None)
            if clean_method and callable(clean_method):
                aws_clean_func = clean_method(value)
                if isawaitable(aws_clean_func):
                    value = await aws_clean_func
                else:
                    value = clean_method(value)
            # url处理
            if f.is_url and domain and value.startswith('/'):
                value = domain + value
            if type(value) == str:
                value = value.strip()

            # 重新设置字段值
            _item.__setitem__(key, value)

        return _item

    def _extract_field(
            self,
            f: Field,
            selector: Selector,
            is_source=False,
    ):
        if not f.xpath_select and not f.css_select and not f.re_select:
            return
        if f.xpath_select and f.css_select:
            raise ValueError(f"{f.__class__.__name__} field: css、xpath选择器只可选择一个.")

        if f.css_select:
            elements = selector.css(f.css_select)
        elif f.xpath_select:
            elements = selector.xpath(f.xpath_select)
        else:
            elements = None

        if elements and is_source:
            return elements

        if elements:
            if f.many:
                results = elements.re(f.re_select) if f.re_select else elements.getall()
            else:
                results = elements.re_first(f.re_select) if f.re_select else elements.get()
        else:
            results = selector.re(f.re_select) if f.re_select else selector.re_first(f.re_select)

        null = [] if f.many else None
        return results or f.default or null



from collections import deque
from functools import wraps


class Middleware:
    """
    Define a middleware to customize the crawler request or response
    eg: middleware = Middleware()
    """

    def __init__(self):
        # request middleware
        self.request_middleware = deque()
        # response middleware
        self.response_middleware = deque()

    def request(self, req_middleware):
        """
        Define a Decorate to be called before a request.
        eg: @middleware.request
        """

        @wraps(req_middleware)
        def register_middleware():
            self.request_middleware.append(req_middleware)
            return req_middleware

        return register_middleware()

    def response(self, resp_middleware):
        """
        Define a Decorate to be called after a response.
        eg: @middleware.response
        """

        @wraps(resp_middleware)
        def register_middleware():
            self.response_middleware.appendleft(resp_middleware)
            return resp_middleware

        return register_middleware()

    def __add__(self, other):
        """
        自定义累加器
        :param other:
        :return:
        """
        # 新的中间件对象
        new_middleware = Middleware()

        # 追加自身的请求、响应中间件
        new_middleware.request_middleware.extend(self.request_middleware)
        new_middleware.request_middleware.extend(other.request_middleware)

        # 追加另一个的请求、响应中间件
        new_middleware.response_middleware.extend(other.response_middleware)
        new_middleware.response_middleware.extend(self.response_middleware)

        # 返回新的中间件对象
        return new_middleware


middleware = Middleware()
